unit HexEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Math;

type
  TGetSelectionEvent = procedure (Sender: TObject; Pos: Integer; var StartPos, len: Integer) of object;
  TGetBufferSelectionEvent = procedure (Sender: TObject; P: PByte; var StartP: PByte; var len: Integer) of object;
  
  THexEdit = class(TCustomControl)
  private
    FBuffer: PByte;
    FBufferSize, FLineCount: Integer;
    FCharWidth, FCharHeight: Integer;
    FCaretX, FCaretY: Integer;
    
    FReadOnly: Boolean;
    FSelectionStart, FSelectionLen: Integer;
    
    FOnGetSelection: TGetSelectionEvent;
    FOnGetBufferSelection: TGetBufferSelectionEvent;
    
    procedure CaretMoved;
    function IsXYInBound(X, Y: Integer; bLoose: Boolean): Boolean;
    procedure BufferPosToCaretXY(Pos: Integer; var X, Y: Integer);
    procedure BufferPosToBufferXY(Pos: Integer; var X, Y: Integer);
    
    function GetCaretBufferPos: Integer;
    procedure SetCaretBufferPos(Value: Integer);
    function GetSelBufferStart: PByte;
    procedure SetSelBufferStart(Value: PByte);
    
    function GetCaretBufferByte: Byte;
    procedure SetCaretBufferByte(Value: Byte);
    procedure SetSelectionLen(Value: Integer);
    
    procedure ShowCorrespondingSelection;
    
    procedure WMSetFont(var message: TWMSetFont); message WM_SETFONT;
    procedure WMSetFocus(var message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMLButtonDown(var message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMGetDlgCode(var message: TWMNoParams); message WM_GETDLGCODE;
  protected
    procedure Paint; override;
    procedure DrawBufferContents;
    procedure DrawBufferLine(LineNo, CharNo, CharEnd: Integer; bSelection: Boolean);
    procedure DrawSelection;
    
    function GetBufferString(LineNo, iPart: Integer): string;
    
    property CaretBufferByte: Byte read GetCaretBufferByte write SetCaretBufferByte;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    
    procedure SetBuffer(Buffer: PByte; BufferSize: Integer);
    procedure AdjustSize;
    
    property SelStart: Integer read GetCaretBufferPos write SetCaretBufferPos;
    property SelLength: Integer read FSelectionLen write SetSelectionLen;
    property SelBufferStart: PByte read GetSelBufferStart write SetSelBufferStart;
  published
    property Align;
    property Color;
    property Ctl3D;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property PopupMenu;
    property Visible;
    
    property readonly: Boolean read FReadOnly write FReadOnly;
    
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;

    property OnGetSelection: TGetSelectionEvent read FOnGetSelection write FOnGetSelection;
    property OnGetBufferSelection: TGetBufferSelectionEvent read FOnGetBufferSelection write FOnGetBufferSelection;
  end;
  
const
  BYTE_PER_LINE = 16;
  
procedure Register;

implementation

constructor THexEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  
  Color := clWhite;
  Font.name := 'Courier New';
  Font.Size := 8;
  Font.Pitch := fpFixed;
  
  FBufferSize := 0;
  FBuffer := nil;
  
  FSelectionStart := 0;
  FSelectionLen := 0;
  
  TabStop := True;
  Visible := True;
end;

destructor THexEdit.Destroy;
begin
  
  
  inherited Destroy;
end;


procedure THexEdit.Paint;
begin
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := Color;
  Canvas.FillRect(Canvas.ClipRect);
  
  DrawBufferContents;
  if FReadOnly then DrawSelection;
end;

function THexEdit.GetBufferString(LineNo, iPart: Integer): string;
var
  CharNo, CharNum, I: Integer;
  S                 : string;
  P                 : PByte;
begin
  CharNo := LineNo * BYTE_PER_LINE;
  if CharNo >= FBufferSize then Exit;
  CharNum := Min(FBufferSize - CharNo, BYTE_PER_LINE);

  case iPart of
    0: Result := Format('%-.4X',[CharNo]);

    1, 2:
    begin
      Result := '';
      P := PByte(Integer(FBuffer) + CharNo);
      for I := 1 to CharNum do
      begin
        if iPart = 1 then
          Result := Result + Format('%-.2X ',[P^])
        else
          Result := Result + PChar(P)^;

        Inc(P);
      end;
      if (CharNum > 0) and (iPart = 1) then
        Delete(Result, Length(Result), 1);
    end;
  end;
end;

procedure THexEdit.DrawBufferLine(LineNo, CharNo, CharEnd: Integer; bSelection: Boolean);
begin
  if LineNo >= FLineCount then Exit;

  with Canvas do
  begin
    Font.Color := clBlue;
    TextOut(0, LineNo * FCharHeight, GetBufferString(LineNo, 0));
    Pen.Color := clLtGray;
    MoveTo(5 * FCharWidth, LineNo * FCharHeight);
    LineTo(5 * FCharWidth, (LineNo + 1) * FCharHeight);
    
    Font.Color := clBlack;
    if bSelection then
    begin
      if Focused then
        Brush.Color := clLime
      else
        Brush.Color := clLtGray;
    end else
      Brush.Color := Color;
    
    if CharEnd = -1 then CharEnd := BYTE_PER_LINE - 1;
    Inc(CharEnd);
    TextOut((6 + CharNo * 3) * FCharWidth, LineNo * FCharHeight, Copy(GetBufferString(LineNo, 1), CharNo * 3 + 1, CharEnd * 3 - (CharNo * 3 + 1))); // ̫@ӰѼƴ@OFLXť
    
    MoveTo((BYTE_PER_LINE * 3 + 6) * FCharWidth, LineNo * FCharHeight);
    LineTo((BYTE_PER_LINE * 3 + 6) * FCharWidth, (LineNo + 1) * FCharHeight);
    
    //    Font.Color := clBlue;
    TextOut((BYTE_PER_LINE * 3 + 7 + CharNo) * FCharWidth, LineNo * FCharHeight, Copy(GetBufferString(LineNo, 2), CharNo + 1, CharEnd - CharNo));

    Brush.Color := Color;
  end;
end;

procedure THexEdit.DrawBufferContents;
var
  I: Integer;
begin
  for I := 0 to FLineCount - 1 do
    DrawBufferLine(I, 0, - 1, False);
end;

procedure THexEdit.DrawSelection;
var
  I             : Integer;
  SX, SY, EX, EY: Integer;
begin
  if FSelectionLen = 0 then Exit;
  
  BufferPosToBufferXY(FSelectionStart, SX, SY);
  BufferPosToBufferXY(FSelectionStart + FSelectionLen - 1, EX, EY);
  
  if SY = EY then
    DrawBufferLine(SY, SX, EX, True)
  else
  begin
    DrawBufferLine(SY, SX, - 1, True);
    for I := SY + 1 to EY - 1 do
      DrawBufferLine(I, 0, - 1, True);
    DrawBufferLine(EY, 0, EX, True);
  end;
end;

procedure THexEdit.SetBuffer(Buffer: PByte; BufferSize: Integer);
begin
  FBuffer := Buffer;
  FBufferSize := BufferSize;
  FLineCount := FBufferSize div BYTE_PER_LINE;
  if FBufferSize mod BYTE_PER_LINE <> 0 then Inc(FLineCount);
  
  FCaretX := 6;
  FCaretY := 0;
  
  FSelectionStart := 0;
  FSelectionLen := 0;
  
  Invalidate;
end;

procedure THexEdit.WMSetFont(var message: TWMSetFont);
var
  TM: TTextMetric;
begin
  if Font.Pitch <> fpFixed then Font.Pitch := fpFixed;
  
  Canvas.Font.Handle := CreateFont(Font.Height, 0, 0, 0, FW_NORMAL, 0, 0, 0, 0, 0, $40, DEFAULT_QUALITY, DEFAULT_PITCH, PChar(Font.name));
  GetTextMetrics(Canvas.Handle, TM);
  FCharWidth := TM.tmAveCharWidth;
  FCharHeight := TM.tmHeight + TM.tmExternalLeading;
  
  Invalidate;
end;

procedure THexEdit.WMSetFocus(var message: TWMSetFocus);
begin
  CreateCaret(Handle, 0, 2, FCharHeight);
  if not FReadOnly and (FBufferSize <> 0) then
  begin
    ShowCaret(Handle);
    CaretMoved;
  end;
  if FReadOnly and (FSelectionLen > 0) then Invalidate;
end;

procedure THexEdit.WMKillFocus(var message: TWMKillFocus);
begin
  if FReadOnly and (FSelectionLen > 0) then Invalidate;
  HideCaret(Handle);
  DestroyCaret;
end;

function THexEdit.IsXYInBound(X, Y: Integer; bLoose: Boolean): Boolean;
begin
  Result := False;
  
  if Y >= FLineCount then Exit
  else
  begin
    if ((X >= 6) and (X < 6 + BYTE_PER_LINE * 3) and (bLoose or ((X - 6) mod 3 <> 2))) and
      ((Y < FLineCount - 1) or (X < 6 + (FBufferSize mod BYTE_PER_LINE) * 3)) then Result := True;
  end;
end;

procedure THexEdit.WMLButtonDown(var message: TWMLButtonDown);
var
  X, Y: Integer;
begin
  if (csDesigning in ComponentState) then Exit;
  
  X := message.XPos div FCharWidth;
  Y := message.YPos div FCharHeight;
  
  if not IsXYInBound(X, Y, FReadOnly) then Exit;
  
  FCaretX := X;
  FCaretY := Y;
  
  if FReadOnly then
  begin
    ShowCorrespondingSelection;
  end else
  begin
    CaretMoved;
  end;
  
  if not Focused then SetFocus;
end;

procedure THexEdit.CaretMoved;
begin
  SetCaretPos(FCaretX * FCharWidth, FCaretY * FCharHeight);
end;

function THexEdit.GetCaretBufferPos: Integer;
begin
  Result := FCaretY * BYTE_PER_LINE + (FCaretX - 6) div 3;
end;

procedure THexEdit.SetCaretBufferPos(Value: Integer);
begin
  if (Value >= 0) and (Value < FBufferSize) then
  begin
    FSelectionStart := Value;
    BufferPosToCaretXY(Value, FCaretX, FCaretY);
    CaretMoved;
  end;
end;

function THexEdit.GetSelBufferStart: PByte;
begin
  Result := PByte(Integer(FBuffer) + GetCaretBufferPos);
end;

procedure THexEdit.SetSelBufferStart(Value: PByte);
begin
  SetCaretBufferPos(Integer(Value) - Integer(FBuffer));
end;


function THexEdit.GetCaretBufferByte: Byte;
begin
  Result := PByte(Integer(FBuffer) + GetCaretBufferPos)^;
end;

procedure THexEdit.SetCaretBufferByte(Value: Byte);
begin
  PByte(Integer(FBuffer) + GetCaretBufferPos)^ := Value;
end;

procedure THexEdit.BufferPosToCaretXY(Pos: Integer; var X, Y: Integer);
begin
  X := Pos mod BYTE_PER_LINE * 3 + 6;
  Y := Pos div BYTE_PER_LINE;
end;

procedure THexEdit.BufferPosToBufferXY(Pos: Integer; var X, Y: Integer);
begin
  X := Pos mod BYTE_PER_LINE;
  Y := Pos div BYTE_PER_LINE;
end;

procedure THexEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if FReadOnly or (FBufferSize = 0) then Exit;
  
  inherited KeyDown(Key, Shift);
  
  case Key of
    VK_HOME:
    begin
      FCaretX := 6;
    end;
    
    VK_END:
    begin
      if FCaretY < FLineCount - 1 then
        FCaretX := BYTE_PER_LINE * 3 + 4
      else
        FCaretX := (FBufferSize mod BYTE_PER_LINE) * 3 + 4;
    end;
    
    VK_PRIOR:
    begin
      FCaretY := 0;
    end;
    
    VK_NEXT:
    begin
      FCaretY := FLineCount - 1;
      Perform(WM_KEYDOWN, VK_END, 0);
    end;
    
    VK_UP:
    begin
      if FCaretY > 0 then Dec(FCaretY);
    end;
    
    VK_DOWN:
    begin
      if (FCaretY < FLineCount - 1) then
        Inc(FCaretY);
      if not IsXYInBound(FCaretX, FCaretY, False) then Perform(WM_KEYDOWN, VK_END, 0);
    end;
    
    VK_LEFT:
    begin
      if (FCaretX > 6) then
      begin
        Dec(FCaretX);
        if (FCaretX - 6) mod 3 = 2 then Dec(FCaretX);
      end else if (FCaretY > 0) then
      begin
        Perform(WM_KEYDOWN, VK_UP, 0);
        Perform(WM_KEYDOWN, VK_END, 0);
      end;
    end;
    
    VK_RIGHT:
    begin
      if (((FCaretY = FLineCount - 1) and (FCaretX < (FBufferSize mod BYTE_PER_LINE) * 3 + 4))
        or ((FCaretY <> FLineCount - 1) and (FCaretX < BYTE_PER_LINE * 3 + 4))) then
      begin
        Inc(FCaretX);
        if (FCaretX - 6) mod 3 = 2 then Inc(FCaretX);
      end else if (FCaretY < FLineCount - 1) then
      begin
        Perform(WM_KEYDOWN, VK_DOWN, 0);
        Perform(WM_KEYDOWN, VK_HOME, 0);
      end;
    end;
  end;
  
  CaretMoved;
end;

procedure THexEdit.KeyPress(var Key: Char);
var
  S    : string;
  B, B1: Byte;  
begin
  if FReadOnly or (FBufferSize = 0) then Exit;
  
  inherited KeyPress(Key);
  
  case Key of
    '0'..'9', 'a'..'f', 'A'..'F':
    begin
      if Key in ['0'..'9'] then
        B1 := Ord(Key) - Ord('0')
      else if Key in ['a'..'f'] then
        B1 := Ord(Key) - Ord('a') + 10
      else B1 := Ord(Key) - Ord('A') + 10;
      
      B := CaretBufferByte;
      if (FCaretX - 6) mod 3 = 0 then
        CaretBufferByte := (B and $0F) or (B1 shl 4)
      else
        CaretBufferByte := (B and $F0) or B1;
      
      Perform(WM_KEYDOWN, VK_RIGHT, 0);
      
      Invalidate;
    end;
  end;
end;

procedure THexEdit.WMGetDlgCode(var message: TWMNoParams);
begin
  message.Result := DLGC_WANTARROWS + DLGC_WANTCHARS;
end;

procedure THexEdit.ShowCorrespondingSelection;
var
  P: PByte;
begin
  if Assigned(FOnGetSelection) then
    FOnGetSelection(Self, GetCaretBufferPos, FSelectionStart, FSelectionLen)
  else if Assigned(FOnGetBufferSelection) then
  begin
    FOnGetBufferSelection(Self, PByte(Integer(FBuffer) + GetCaretBufferPos), P, FSelectionLen);
    FSelectionStart := Integer(P) - Integer(FBuffer);
  end else
  begin
    FSelectionStart := GetCaretBufferPos;
    FSelectionLen := 0;
  end;

  Invalidate;
end;

procedure THexEdit.AdjustSize;
begin
  Perform(WM_SETFONT, 0, 0);
  ClientWidth := FCharWidth * (BYTE_PER_LINE shl 2 + 7);
  ClientHeight := FCharHeight * FLineCount;
end;

procedure THexEdit.SetSelectionLen(Value: Integer);
begin
  FSelectionLen := Value;
  Invalidate;
end;

procedure Register;
begin
  RegisterComponents('Delphi Workshop',[THexEdit]);
end;

end.

